/* 
 * (C) Copyright 1992, ..., 2000 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING in the DOSEMU distribution
 */

#ifndef _EXTERN_H
#define _EXTERN_H

/* some included defines in all header files when
 * this isn't data.c 
 */

#ifndef INIT
#define INIT(x...)
#endif

#ifndef EXTERN
#define EXTERN extern
#endif

#endif

